@@ -1,7 +1,7 @@
 """PAX Module"""
 
-from .flatten_module import flatten_module
-from .mixed_precision import apply_mp_policy
+from .graph_module import GraphModule, InputNode, build_graph_module
+from .mixed_precision import apply_mp_policy, unwrap_mp_policy
 from .module import Module, PaxKind
 from .module_and_value import module_and_value
 from .pure import pure
@@ -15,4 +15,5 @@ from .transforms import (
     update_parameters,
     update_states,
 )
+from .utility_modules import Flattener, LazyModule, ParameterModule, StateModule
 from .utils import assert_structure_equal
