@@ -12,7 +12,6 @@ from enum import Enum
 from types import MappingProxyType
 from typing import (
     Any,
-    Dict,
     Iterable,
     List,
     Mapping,
@@ -21,7 +20,6 @@ from typing import (
     Tuple,
     Type,
     TypeVar,
-    Union,
 )
 
 import jax
@@ -33,13 +31,7 @@ from jax.dtypes import issubdtype as isdt
 # pylint: disable=no-name-in-module
 from jaxlib.xla_extension import CompiledFunction
 
-from .threading_local import (
-    add_mutable_module,
-    allow_mutation,
-    is_deep_copy_enabled,
-    is_inside_pure_function,
-    is_mutable,
-)
+from .threading_local import allow_mutation, is_deep_copy_enabled, is_mutable
 
 T = TypeVar("T", bound="BaseModule")
 M = TypeVar("M")
@@ -83,14 +75,13 @@ class PaxKind(Enum):
 class PaxModuleInfo(NamedTuple):
     """PAX Internal Data Structure."""
 
-    name: Optional[str]
     training: bool
     name_to_kind: Mapping[str, PaxKind]
     default_kind: PaxKind
 
     def __repr__(self) -> str:
         nodes = ", ".join([f"{k}:{v.name}" for k, v in self.name_to_kind.items()])
-        return f"PaxModuleInfo[name={self.name}, training={self.training}, nodes={{{nodes}}}]"
+        return f"PaxModuleInfo[training={self.training}, nodes={{{nodes}}}]"
 
 
 class BaseModuleMetaclass(type):
@@ -99,17 +90,14 @@ class BaseModuleMetaclass(type):
     def __call__(cls: Type[T], *args, **kwargs) -> T:
         module = cls.__new__(cls, *args, **kwargs)  # type: ignore
 
-        # if a module is created inside a `pure` function, it is mutable.
-        if is_inside_pure_function():
-            add_mutable_module(module)
-            cls.__init__(module, *args, **kwargs)
-            module.find_and_register_submodules()
-        else:
-            with allow_mutation(module):
-                cls.__init__(module, *args, **kwargs)
-                module.find_and_register_submodules()
+        with allow_mutation(module):
+            with module._default_kind(module._pax.default_kind):
+                with module._default_kind(PaxKind.MODULE):
+                    cls.__init__(module, *args, **kwargs)
 
         # scan module after initialization for potential bugs
+        module._assert_not_shared_module()
+        module._assert_not_shared_weight()
         module._scan_fields(module.__dict__.keys())
         return module
 
@@ -130,6 +118,10 @@ class BaseModule(metaclass=BaseModuleMetaclass):
 
     _pax: PaxModuleInfo
 
+    @property
+    def pax(self):
+        return self._pax
+
     def __new__(cls: Type[T], *args, **kwargs) -> T:
         """Initialize `_pax` attribute in `__new__` method to avoid
         calling `super().__init__()` in every subclass of Module."""
@@ -140,7 +132,6 @@ class BaseModule(metaclass=BaseModuleMetaclass):
         super(BaseModule, obj).__setattr__(
             "_pax",
             PaxModuleInfo(
-                name=None,
                 training=True,
                 name_to_kind=MappingProxyType(OrderedDict()),
                 default_kind=PaxKind.UNKNOWN,
@@ -150,14 +141,15 @@ class BaseModule(metaclass=BaseModuleMetaclass):
         if obj.__slots__ != ("_pax",):
             raise ValueError("`__slots__` is not supported by PAX modules.")
 
-        # scan class attributes for unregistered modules and ndarray's.
-        obj._scan_fields(obj.__class__.__dict__)
+        # scan class attributes for unregistered modules and ndarrays.
+        obj._scan_fields(obj._class_fields())
 
         return obj
 
-    def __init__(self, name: Optional[str] = None):
-        """Initialize module's name."""
-        super().__setattr__("_pax", self._pax._replace(name=name))
+    def _class_fields(self):
+        for name, value in self.__class__.__dict__.items():
+            if not hasattr(value, "__get__"):  # ignore descriptors
+                yield name
 
     def _assert_mutability(self):
         if not is_mutable(self):
@@ -178,12 +170,7 @@ class BaseModule(metaclass=BaseModuleMetaclass):
         super().__setattr__("_pax", new_info)
 
     def __setattr__(self, name: str, value: Any) -> None:
-        """Whenever a user sets an attribute, we will check the assignment:
-
-        - Setting `_pax` is forbidden.
-        - If `value` is a pytree of modules and `name` is not in `name_to_kind`,
-          its kind will be `PaxKind.MODULE`.
-        """
+        """Setting `_pax` attribute is forbidden."""
         self._assert_mutability()
 
         if name == "_pax":
@@ -192,17 +179,10 @@ class BaseModule(metaclass=BaseModuleMetaclass):
             )
 
         super().__setattr__(name, value)
-        if name not in self._pax.name_to_kind:
-            if self._pax.default_kind != PaxKind.UNKNOWN:
-                self._update_name_to_kind_dict(name, self._pax.default_kind)
-
-            self.find_and_register_submodules()
 
     def __delattr__(self, name: str) -> None:
-        self._assert_mutability()
-        if name in self._pax.name_to_kind:
-            raise ValueError("Cannot delete a pytree attribute.")
-        super().__delattr__(name)
+        del name
+        raise ValueError("Cannot delete a module's attribute.")
 
     def _update_default_kind(self, kind: PaxKind):
         self._assert_mutability()
@@ -210,21 +190,39 @@ class BaseModule(metaclass=BaseModuleMetaclass):
         super(BaseModule, self).__setattr__("_pax", new_pax_info)
 
     @contextmanager
-    def default_kind(self, kind: PaxKind):
-        prev = self._pax.default_kind
+    def _default_kind(self, kind: PaxKind):
+        """Search for any new unregistered pytree attribute
+        at the end of the context and register it with kind `kind`.
+        """
+        fields_before = set(vars(self).keys())
         try:
-            self._update_default_kind(kind)
             yield
         finally:
-            self._update_default_kind(prev)
+            fields_after = vars(self).keys()
+            new_fields = (name for name in fields_after if name not in fields_before)
+            self._find_and_register_pytree(kind, fields=new_fields)
 
     def add_parameters(self):
-        """Add new attributes as trainable parameters"""
-        return self.default_kind(PaxKind.PARAMETER)
+        """Return a context manager that registers new attributes
+        in the scope as trainable parameters.
+
+        >>> def add_param(net):
+        ...     with net.add_parameters():
+        ...         net.weight = jnp.array(0.)
+        ...         net.bias = jnp.array(0.)
+        ...     return net
+        ...
+        >>> net = pax.pure(add_param)(pax.Module())
+        >>> net._pax
+        PaxModuleInfo[training=True, nodes={weight:PARAMETER, bias:PARAMETER}]
+        """
+        return self._default_kind(PaxKind.PARAMETER)
 
     def add_states(self):
-        """Add new attributes as non-trainable states."""
-        return self.default_kind(PaxKind.STATE)
+        """Return a context manager that registers new attributes
+        in the scope as non-trainable states.
+        """
+        return self._default_kind(PaxKind.STATE)
 
     def tree_flatten(
         self,
@@ -264,10 +262,6 @@ class BaseModule(metaclass=BaseModuleMetaclass):
         # pylint: disable=protected-access
         module_dict.update(zip(module._pax.name_to_kind, children))
 
-        # if a module is created inside a `pure` function, it is mutable.
-        if is_inside_pure_function():
-            add_mutable_module(module)
-
         return module
 
     def __init_subclass__(cls):
@@ -303,7 +297,7 @@ class BaseModule(metaclass=BaseModuleMetaclass):
                             f"(type={type(leaf)}, value={leaf})."
                         )
 
-            # Check if a PARAMETER attribute contains non-differentiable ndarray's.
+            # Check if a PARAMETER attribute contains non-differentiable ndarrays.
             if kind == PaxKind.PARAMETER:
                 for leaf in leaves:
                     if hasattr(leaf, "dtype") and not (
@@ -334,9 +328,9 @@ class BaseModule(metaclass=BaseModuleMetaclass):
                 for leaf in leaves:
                     if isinstance(leaf, (np.ndarray, jnp.ndarray)):
                         raise ValueError(
-                            f"Unregistered field `{self}.{name}` ({kind}) contains a ndarray."
-                            f" Consider registering it using `self.set_attribute_kind`"
-                            f" or `self.register_*` methods."
+                            f"Unregistered field `{self.__class__.__name__}.{name}`"
+                            f" ({kind}) contains a ndarray. Consider registering it using"
+                            f" `self.add_parameters()` or `self.add_states()` context managers."
                         )
 
             # Check if an unregistered (or PARAMETER) field contains pax.Module instances
@@ -373,32 +367,56 @@ class BaseModule(metaclass=BaseModuleMetaclass):
         leaves = jax.tree_map(lambda x: (x.shape, x.dtype), leaves)
         return hash((tuple(leaves), treedef))
 
-    def find_and_register_submodules(self):
-        """Find unregistered submodules and register it with MODULE kind."""
+    def _find_and_register_pytree(
+        self, kind: PaxKind, fields: Optional[Iterable[str]] = None
+    ):
+        """Find unregistered pytree attributes and register it with kind `kind`."""
 
-        def all_module_leaves(tree):
-            leaves = jax.tree_flatten(
-                tree, is_leaf=lambda m: isinstance(m, BaseModule)
-            )[0]
-            return len(leaves) > 0 and all(isinstance(m, BaseModule) for m in leaves)
+        if kind == PaxKind.UNKNOWN:
+            return
 
-        for name, value in vars(self).items():
+        if fields is not None:
+            items = ((name, getattr(self, name)) for name in fields)
+        else:
+            items = vars(self).items()
+
+        def is_pytree_of(tree, classes):
+            is_leaf_fn = lambda m: isinstance(m, classes)
+            leaves = jax.tree_flatten(tree, is_leaf=is_leaf_fn)[0]
+            return len(leaves) > 0 and all(is_leaf_fn(m) for m in leaves)
+
+        for name, value in items:
             if name not in self._pax.name_to_kind:
-                if all_module_leaves(value):
-                    self._update_name_to_kind_dict(name, PaxKind.MODULE)
+                if (kind == PaxKind.MODULE and is_pytree_of(value, BaseModule)) or (
+                    kind in [PaxKind.PARAMETER, PaxKind.STATE]
+                    and not _has_module_node(value)
+                    and is_pytree_of(value, (np.ndarray, jnp.ndarray))
+                ):
+                    self._update_name_to_kind_dict(name, kind)
 
     def register_subtree(self, name: str, value, kind: PaxKind):
         """Assign `value` to attribute `name` and register its kind as `kind`."""
         if hasattr(self, name):
             raise RuntimeError("Cannot register an existing attribute")
 
-        self._update_name_to_kind_dict(name, kind)
         setattr(self, name, value)
+        self.set_attribute_kind(**{name: kind})
 
     def set_attribute_kind(self, **kwargs):
-        """Example:
+        """Set PAX kind for one or many attributes.
 
-        >>> self.set_attribute_kind(weight=pax.P, counter=pax.S)
+        Example:
+
+        >>> class M(pax.Module):
+        ...     def __init__(self):
+        ...         super().__init__()
+        ...         self.weight = jnp.array(0.)
+        ...         self.counter = jnp.array(0)
+        ...         self.set_attribute_kind(weight=pax.P, counter=pax.S)
+        ...
+        >>> m = M()
+        >>> print(m._pax.name_to_kind)
+        OrderedDict([('weight', <PaxKind.PARAMETER: 2>), ('counter', <PaxKind.STATE: 1>)])
         """
         for name, kind in kwargs.items():
             if not hasattr(self, name):
@@ -408,7 +426,16 @@ class BaseModule(metaclass=BaseModuleMetaclass):
     def apply(self: T, apply_fn) -> T:
         """Apply a function to all submodules.
 
-        **Note**: this function returns a transformed copy of the module.
+        >>> def print_param_count(mod):
+        ...     count = sum(jax.tree_leaves(jax.tree_map(jnp.size, mod)))
+        ...     print(f"{count}\t{mod}")
+        ...     return mod
+        ...
+        >>> net = pax.nn.Sequential(pax.nn.Linear(1, 1), jax.nn.relu)
+        >>> net = net.apply(print_param_count)
+        2 Linear(in_dim=1, out_dim=1, with_bias=True)
+        0 Lambda(relu)
+        2 Sequential
 
         Arguments:
             apply_fn: a function which inputs a module and outputs a transformed module.
@@ -444,55 +471,57 @@ class BaseModule(metaclass=BaseModuleMetaclass):
         submods, _ = jax.tree_flatten(module_subtrees, is_leaf=is_module)
         return [v for v in submods if is_module(v)]
 
-    def summary(self, return_list: bool = False) -> Union[str, List[str]]:
-        """This is the default summary method.
-
-        Arguments:
-            return_list: return a list of lines instead of a joined string.
-
-
-        Example:
-
-        >>> net = pax.nn.Sequential(pax.nn.Linear(2, 3), jax.nn.relu, pax.nn.Linear(3, 4))
-        >>> print(net.summary())
-        Sequential
-        ├── Linear[in_dim=2, out_dim=3, with_bias=True]
-        ├── x => relu(x)
-        └── Linear[in_dim=3, out_dim=4, with_bias=True]
-        """
-
-        output = [self.__repr__()]
-        if output[0] is None:
+    def _assert_not_shared_module(self):
+        """Shared module is not allowed."""
+        shared_module = _find_shared_module(self)
+        if shared_module is not None:
             raise ValueError(
-                f"The `{self.__class__}.__repr__` method returns a `None` value."
+                f"The module `{shared_module}` is shared between two nodes of the pytree.\n"
+                f"This is not allowed to prevent potential silence bugs."
             )
-        submodules: List[BaseModule] = self.submodules()
 
-        def indent(lines: List[str], start_string) -> List[str]:
-            return [start_string + l for l in lines]
+    def _assert_not_shared_weight(self):
+        """Shared weight is not allowed."""
+        leaves = jax.tree_leaves(self)
+        leaf_ids = set()
+        for leaf in leaves:
+            if id(leaf) in leaf_ids:
+                raise ValueError(
+                    f"Detected a shared ndarray. This is not allowed.\n"
+                    f"Shape={leaf.shape}\n"
+                    f"Dtype={leaf.dtype}\n"
+                    f"Value={leaf}",
+                )
+            leaf_ids.add(id(leaf))
 
-        for i, module in enumerate(submodules):
-            lines = module.summary(return_list=True)
-            if i + 1 < len(submodules):  # middle submodules
-                indented_lines = indent(lines[:1], "├── ") + indent(lines[1:], "│   ")
-            else:  # last submodule
-                indented_lines = indent(lines[:1], "└── ") + indent(lines[1:], "    ")
-            output.extend(indented_lines)
 
-        if return_list:
-            return output
-        else:
-            return "\n".join(output)
+def _find_shared_module(module: BaseModule):
+    """Find shared module.
 
-    def _repr(self, info: Optional[Dict[str, Any]] = None) -> str:
-        name = f"({self._pax.name}) " if self._pax.name is not None else ""
-        cls_name = self.__class__.__name__
-        if info is None:
-            return f"{name}{cls_name}"
-        else:
-            lst_info = [f"{k}={v}" for (k, v) in info.items() if v is not None]
-            str_info = ", ".join(lst_info)
-            return f"{name}{cls_name}[{str_info}]"
+    - Return the first module that is shared by two nodes of the pytree.
+    - Return `None` if there is no shared module.
+    """
 
-    def __repr__(self) -> str:
-        return self._repr()
+    def _get_all_modules(mod: BaseModule, lst: List):
+        lst.append(mod)
+        for m in mod.submodules():
+            _get_all_modules(m, lst)
+
+    mods = []
+    _get_all_modules(module, mods)
+    module_ids = set()
+    for m in mods:
+        if id(m) in module_ids:
+            return m
+        module_ids.add(id(m))
+
+    return None
+
+
+def _has_module_node(value):
+    leaves = jax.tree_flatten(value, is_leaf=lambda x: isinstance(x, BaseModule))[0]
+    for leaf in leaves:
+        if isinstance(leaf, BaseModule):
+            return True
+
+    return False
