@@ -3,7 +3,7 @@
 import functools
 import gc
 from types import MethodType
-from typing import Callable, Optional, Sequence, Union
+from typing import Optional, Sequence, Union
 
 import jax
 
@@ -14,11 +14,17 @@ from .threading_local import (
     get_rng_state,
     set_rng_state,
 )
-from .utils import get_modules
+
+
+def _get_modules(tree):
+    "Return a list of modules in the pytree `tree`."
+    modules = jax.tree_flatten(tree, is_leaf=lambda x: isinstance(x, BaseModule))[0]
+    modules = [m for m in modules if isinstance(m, BaseModule)]
+    return modules
 
 
 def _get_all_submodules(value):
-    submods = get_modules(value)
+    submods = _get_modules(value)
     out = list(submods)
     for mod in submods:
         out.extend(_get_all_submodules(mod.submodules()))
@@ -26,10 +32,10 @@ def _get_all_submodules(value):
 
 
 def pure(
-    func: Callable,
+    func,
     static_argnums: Optional[Union[int, Sequence[int]]] = None,
     check_leaks: bool = True,
-) -> Callable:
+):
     """Make a function pure by copying the inputs.
 
     Any modification on the copy will not affect the original inputs.
@@ -40,9 +46,10 @@ def pure(
 
     >>> f = pax.nn.Linear(3,3)
     >>> f.a_list = []
-    [...]
+    Traceback (most recent call last):
+      ...
     ValueError: Cannot modify a module in immutable mode.
-    Please do this computation inside a @pax.pure function.
+    Please do this computation inside a function decorated by `pax.pure`.
     >>>
     >>> @pax.pure
     ... def add_list(m):
@@ -77,7 +84,7 @@ def pure(
 
     @functools.wraps(func)
     def _f(*args, **kwargs):
-        _ = [m.scan_bugs() for m in get_modules((func, args, kwargs))]
+        _ = [m.scan_bugs() for m in _get_modules((func, args, kwargs))]
 
         # support calling method
         if isinstance(func, MethodType):
@@ -99,33 +106,33 @@ def pure(
         args = tuple(args)
 
         def no_leak_func(*args, **kwargs):
-            gc.collect()
-            args = list(args)
-            for i in static_argnums:
-                args[i] = args_copy[i]
-            args = tuple(args)
-            set_rng_state(rng_state)
-            out = unbound_func(*args, **kwargs)
-            set_rng_state(rng_state)
-            gc.collect()
-            return out
-
-        def _run(args, kwargs, eval_shape: bool = False):
             args, kwargs = _deepcopy((args, kwargs))
             modules = _get_all_submodules((args, kwargs))
             with allow_mutation(modules):
-                if eval_shape:
-                    out = jax.eval_shape(no_leak_func, *args, **kwargs)
-                else:
-                    out = no_leak_func(*args, **kwargs)
+                gc.collect()
+                args = list(args)
+                for i in static_argnums:
+                    args[i] = args_copy[i]
+                args = tuple(args)
+                set_rng_state(rng_state)
+                out = unbound_func(*args, **kwargs)
+                set_rng_state(rng_state)
+                gc.collect()
                 return out
 
+        def _run(args, kwargs, eval_shape: bool = False):
+            if eval_shape:
+                out = jax.eval_shape(no_leak_func, *args, **kwargs)
+            else:
+                out = no_leak_func(*args, **kwargs)
+            return out
+
         with jax.check_tracer_leaks(check_leaks):
-            # leak check
-            _run(args, kwargs, eval_shape=True)
+            if check_leaks:
+                _run(args, kwargs, eval_shape=True)
             # real run
             out = _run(args, kwargs, eval_shape=False)
-        _ = [m.scan_bugs() for m in get_modules(out)]
+        _ = [m.scan_bugs() for m in _get_modules(out)]
         return out
 
     return _f
