@@ -4,6 +4,7 @@ Manage thread local states
 
 import logging
 import threading
+import weakref
 from contextlib import contextmanager
 from typing import Any, Optional, Tuple, Union
 
@@ -19,14 +20,13 @@ class PaxThreadingLocalState(threading.local):
 
     __slots__ = [
         "_enable_deep_copy",
-        "_inside_pure_function",
-        "_mutable_module_id_list",
+        "_mutable_module_ref_list",
         "_rng_seed",
         "_rng_key",
     ]
     _enable_deep_copy: bool
-    _inside_pure_function: bool
-    _mutable_module_id_list: Tuple[int, ...]
+    _mutable_module_ref_list: Tuple[weakref.ReferenceType, ...]
+    _mutable_module_level: jax.core.Sublevel
     _rng_seed: Optional[int]
     _rng_key: Optional[KeyArray]
 
@@ -34,26 +34,35 @@ class PaxThreadingLocalState(threading.local):
         super().__init__()
 
         self._enable_deep_copy = False
-        self._inside_pure_function = False
-        self._mutable_module_id_list = ()
+        self._mutable_module_ref_list = ()
+        self._mutable_module_level = jax.core.cur_sublevel()
         self._rng_seed = None
         self._rng_key = None
 
-    def is_inside_pure_function(self):
-        """are we inside a function decorated by `pax.pure`"""
-        return self._inside_pure_function
-
     def is_deep_copy_enabled(self):
         """use deepcopy to copy modules"""
         return self._enable_deep_copy
 
     def add_mutable_module(self, module):
         """add `module` to mutable list"""
-        self._mutable_module_id_list = (id(module),) + self._mutable_module_id_list
+        self._mutable_module_ref_list = (
+            weakref.ref(module),
+            *self._mutable_module_ref_list,
+        )
 
     def is_mutable(self, module):
         """Is `module` mutable?"""
-        return id(module) in self._mutable_module_id_list
+
+        # cannot modify a module whose level of abstraction
+        # is lower than the current level
+        if self._mutable_module_level < jax.core.cur_sublevel():
+            return False
+
+        for ref in self._mutable_module_ref_list:
+            if module is ref():
+                return True
+
+        return False
 
     @contextmanager
     def enable_deep_copy(self):
@@ -70,18 +79,18 @@ class PaxThreadingLocalState(threading.local):
         r"""A context manager that turns on mutability."""
         if not isinstance(modules, (tuple, list)):
             modules = (modules,)
-        modules = tuple(id(mod) for mod in modules)
+        modules = tuple(weakref.ref(mod) for mod in modules)
 
-        prev = self._mutable_module_id_list
-        prev_inside = self._inside_pure_function
+        prev = self._mutable_module_ref_list
+        prev_abstraction_level = self._mutable_module_level
         prev_rng_state = self.get_rng_state()
         try:
-            self._inside_pure_function = True
-            self._mutable_module_id_list = modules
+            self._mutable_module_ref_list = modules
+            self._mutable_module_level = jax.core.cur_sublevel()
             yield
         finally:
-            self._mutable_module_id_list = prev
-            self._inside_pure_function = prev_inside
+            self._mutable_module_ref_list = prev
+            self._mutable_module_level = prev_abstraction_level
             self.set_rng_state(prev_rng_state)
 
     def seed_rng_key(self, seed: int) -> None:
@@ -141,7 +150,6 @@ allow_mutation = PAX_STATE.allow_mutation
 enable_deep_copy = PAX_STATE.enable_deep_copy
 get_rng_state = PAX_STATE.get_rng_state
 is_deep_copy_enabled = PAX_STATE.is_deep_copy_enabled
-is_inside_pure_function = PAX_STATE.is_inside_pure_function
 is_mutable = PAX_STATE.is_mutable
 next_rng_key = PAX_STATE.next_rng_key
 seed_rng_key = PAX_STATE.seed_rng_key
