@@ -5,16 +5,9 @@ from unittest import TestCase
 
 import jax
 
-from .module import Module
+from .base import BaseModule
 
-T = TypeVar("T", bound=Module)
-
-
-def get_modules(tree):
-    "Return a list of modules in the pytree `tree`."
-    modules = jax.tree_flatten(tree, is_leaf=lambda x: isinstance(x, Module))[0]
-    modules = [m for m in modules if isinstance(m, Module)]
-    return modules
+T = TypeVar("T", bound=BaseModule)
 
 
 def assert_structure_equal(tree_a: T, tree_b: T):
@@ -26,7 +19,7 @@ def assert_structure_equal(tree_a: T, tree_b: T):
         return True
 
     def check(subtree_a, subtree_b):
-        if isinstance(subtree_a, Module) and isinstance(subtree_b, Module):
+        if isinstance(subtree_a, BaseModule) and isinstance(subtree_b, BaseModule):
             assert_structure_equal(subtree_a, subtree_b)
 
     has_error = False
@@ -35,7 +28,7 @@ def assert_structure_equal(tree_a: T, tree_b: T):
             check,
             tree_a,
             tree_b,
-            is_leaf=lambda x: isinstance(x, Module)
+            is_leaf=lambda x: isinstance(x, BaseModule)
             and x is not tree_a
             and x is not tree_b,
         )
